#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import argparse
import shlex
from pathlib import Path
from queue import Queue

from p1st.common import Message1, Message2, get_chunk_path, get_inform1_path, get_inform2_path
from p1st.exec_consume_chunks import execute_consume_chunks
from p1st.repeat import repeat_until_successful


def main():


    args = parse_args()
    #
    ssh_source: str = args.ssh_source
    source_chunk_dir: Path = args.source_chunk_dir
    target_chunk_dir: Path = args.chunk_dir
    target_path: Path = args.target_path

    target_path.parent.mkdir(parents=True, exist_ok=True)

    command = ['btrfs', 'receive', str(target_path.parent)]

    def handle_chunks(queue_put: Queue.put):
        chunk_no = 1
        while True:
            target_chunk_path = get_chunk_path(target_chunk_dir, target_path, chunk_no)
            source_chunk_path = get_chunk_path(source_chunk_dir, target_path, chunk_no)
            in_socket: Path = get_inform2_path(source_chunk_dir, target_path)

            user_socket = target_chunk_dir.parent.joinpath(f'{target_path.name}.SOCKET.user')

            inform_path = get_inform1_path(source_chunk_path)
            inform_cmd = ['ssh',
                          ssh_source,
                          f'cat {shlex.quote(str(inform_path))} && rm {shlex.quote(str(inform_path))}']
            messages = [Message1.OK.value, Message1.EOF.value]
            rsync_cmd = ['rsync',
                         f'{ssh_source}:{source_chunk_path}',
                         f'{target_chunk_path}']
            inform2_cmd = ['ssh',
                           ssh_source,
                           f"printf '{Message2.OK_STR.value}' | nc -U {shlex.quote(str(in_socket))}"]

            # Wait until next chunk can be transferred from sending side.
            #
            # It can happen,that the sending side has not yet saved the next chunk.
            # In that case `cat ...` will fail.
            # We will retry 3 times, each time after 5 seconds sleep.
            #
            # If all 4 `cat ...` attempts failed, we wait until the user has inspected the problem.
            msg = repeat_until_successful(inform_cmd, user_socket, retries=3, retry_delay_seconds=5)
            if msg not in messages:
                raise ValueError(f'Invalid message: {msg}')
            last_chunk = msg == Message1.EOF.value

            # Transfer chunk.
            repeat_until_successful(rsync_cmd, user_socket)

            # Add chunk path to queue.
            queue_put((target_chunk_path, last_chunk))

            # Inform sending side about successful transfer.
            repeat_until_successful(inform2_cmd, user_socket)

            if last_chunk:
                break
            chunk_no += 1

    returncode = execute_consume_chunks(
        command=command,
        handle_chunks=handle_chunks,
    )
    exit(returncode)


def parse_args():
    parser = argparse.ArgumentParser(prog='btrfs-receive-chunks')

    parser.add_argument('--chunk-dir',
                        help='Chunks are saved in this directory. '
                             'Defaults to parent directory of SUBVOLUME.',
                        dest='chunk_dir',
                        type=Path,
                        metavar='CHUNK_DIR',
                        default=None,
                        )

    parser.add_argument('ssh_source',
                        help='Hostname of source computer; as configured in ~/.ssh/config.',
                        metavar='SSH_SOURCE'
                        )

    parser.add_argument('source_chunk_dir',
                        help='Chunks are saved in this directory on SSH_SOURCE. '
                             'Must be an absolute path.',
                        type=Path,
                        metavar='SOURCE_CHUNK_DIR',
                        )

    parser.add_argument('target_path',
                        help='Path where the subvolume will be created. '
                             'The last component of the path '
                             'must be equal to the name of the subvolume on the sending side.',
                        type=Path,
                        metavar='SUBVOLUME'
                        )

    args = parser.parse_args()

    # Make all paths absolute. Set default values.
    args.target_path = args.target_path.absolute()
    if args.chunk_dir:
        args.chunk_dir = args.chunk_dir.absolute()
    else:
        args.chunk_dir = args.target_path.absolute().parent
    if not args.source_chunk_dir.is_absolute():
        raise ValueError(f'SOURCE_CHUNK_DIR must be absolute')

    return args


if __name__ == '__main__':
    main()
